/**
	@file oemstar_gps.c
	
	Control code for the NovaTel OEMStar GPS board
  
  Example hardware setup for communicating with OEMStar over UART0:
  @code
  // *******************************************************************************
	// Initialize B2A2 board for communication over UART0
	// *******************************************************************************
  U0LCR=0x83;  //Step 1: DLAB=1, enable baud rate adjustments
  //Set to 8 data bits, 1 stop bit, no parity
	//Step 2: Set baud rate divisor and fractional baud rate divisors
  U0DLL=25;		//Baud rate divisor for 115200 kbaud, with (1+0.3) fractional divider 
  U0DLM=0;		//and peripheral clock = 60MHz.
  U0FDR = 3;		// DIVADDVAL = 3
  U0FDR |= 10<<4;	// MULTVAL = 10
  //U0FCR = 0x81;    	// Enable receive and transmit FIFO; interrupt on receive at 8 bytes.
  U0FCR = 0x1;		// Enable FIFO
  U0FCR = 0x87;		//Clear FIFOs and set to generate interrupt when RX FIFO reaches 8 bytes.
  U0MCR = 0;	   	// Loopback mode, CTS and RTS flow control disabled.
  U0ACR = 0;	   	// Autobaud disabled
  U0TER = 0x80;	   	// Transmit enabled
  U0LCR &= ~(1<<7); // Step 3: DLAB=0, disable baud rate adjustments
  PINSEL0 |= 1<<16;    // UART1 TXD active on MCU pin P0[8]
  PINSEL0 &= ~(1<<17); //
  PINSEL0 |= 1<<18;    // UART1 RXD	active on MCU pin P0[9]
  PINSEL0 &= ~(1<<19); //
  U0IER = (1<<0); // enable RBR interrupts
  U0ACR = (1<<8) | (1<<9); //clear the corresponding interrupt in the U1IIR
  @endcode
  
  @author Nicolas Williamson
  @author Jason Cortell
  @date March 2011
	
*/

#include <includes.h>

//Define GPS command structures.
GPS_ANTENNAPOWER gps_antennapower_command;
GPS_CLOCKADJUST gps_clockadjust_command;
GPS_COM gps_com_command;
GPS_ECUTOFF gps_ecutoff_command;
GPS_FIX gps_fix_command;
GPS_FRESET gps_freset_command;
GPS_LOG gps_log_command;
GPS_PDPFILTER gps_pdpfilter_command;
GPS_PDPMODE gps_pdpmode_command;
GPS_PSRVELOCITYTYPE gps_psrvelocitytype_command;
GPS_RESET gps_reset_command;
GPS_SBASCONTROL gps_sbascontrol_command;
GPS_SAVECONFIG gps_saveconfig_command;
GPS_SETIONOTYPE gps_setionotype_command;

//Define GPS command response:
GPS_COMMAND_RESPONSE gps_response;

//Define GPS received data structure; must include all desired logs.
__packed struct GPS_RECEIVED_DATA
{
  GPS_PDPVEL pdpvel;
  GPS_PDPXYZ pdpxyz;
} gps_received_data; 

//Global variables for UART0 software receive buffer
#define GPS_RX_BUFFER_SIZE 128    //Buffer size must be a power of two!
volatile unsigned char gps_read_buffer[GPS_RX_BUFFER_SIZE];	//FIFO read buffer for data from UART0
volatile unsigned char gps_rb_index1 = 1;			//Points to location of next open spot in buffer
volatile unsigned char gps_rb_index2 = 0;			//Points to location of oldest unused rec. byte
//If Index1 = (Index2 + 1) & (GPS_RX_BUFFER_SIZE - 1), buffer is empty.
//Buffer is full when Index1 = Index2

//Global variables for UART0 software transmit buffer
#define GPS_TX_BUFFER_SIZE 128    //Buffer size must be a power of two!
volatile unsigned char gps_trans_buffer[GPS_TX_BUFFER_SIZE];		//FIFO transmit buffer for data to UART
volatile unsigned char gps_tb_index1 = 0;			//Points to location of next open spot in buffer
volatile unsigned char gps_tb_index2 = 0;			//Points to location of oldest byte not yet transmitted
//If Index1 = Index2 - 1 (with appropriate wraparound), buffer is full.
//Buffer is empty when Index1 = Index2


//Initialize GPS for desired functions and operations. Because of the time required for the GPS to be fully initialized
//after power-up, this init function should be called not from software setup (as with most of the init functions),
//but repeatedly during normal operation. This also allows the initialization process to be repeated later, if necessary.
//State variables are used to indicate which initialization steps have completed successfully; these states can be reset if
//reinitialization is required.
void gps_init(void)
{
static unsigned char log_pointers_state = 1;
static unsigned char antennapower_state = 3, clockadjust_state = 3, com_state = 3, ecutoff_state = 3, fix_state = 3, freset_state = 3;
static unsigned char log_state = 3, pdpfilter_state = 3, pdpmode_state = 3, psrvelocitytype_state = 3, reset_state = 3, sbascontrol_state = 3;
static unsigned char saveconfig_state = 3, setionotype_state = 3;
static unsigned long timeout = 0;

//Init step 1
if (antennapower_state)
{
//erase command_response structure; reset timeout counter
  if (antennapower_state == 3)
  {
    timeout = 0;  //reset timeout variable

  }

//send command
//wait for successful command response; restart state machine if needed
}

}


/**
  The interrupt service routine for UART0.
  Stores incoming data to rx software buffer, sends out data
  from tx software buffer.
*/
void gps_isr(void) __irq
{	
  unsigned long interrupt_id, line_status;
	unsigned short i;

	interrupt_id = U0IIR;
	
	switch (interrupt_id & 0xE) //Look at bits 1 - 3 of U0IIR
	{
	case 4:	//receive data available (>=8 bytes in hardware receive FIFO, as set in U0FCR)
	  for (i=0;i<8;i++)
	  {
	    line_status = U0LSR;
	    if (gps_rb_index1 != gps_rb_index2)			  //software receive buffer not full
	    {
	      gps_read_buffer[gps_rb_index1] = U0RBR;	//Put new data in next buffer location
        if (gps_rb_index1 == GPS_RX_BUFFER_SIZE - 1)
        {
          gps_rb_index1 = 0;  //Roll over gps_rb_index1
        }
        else
        {
          gps_rb_index1++;    //Increment gps_rb_index1
        }
      }
      else //Software receive buffer overflow
      {
         error_occurred_irq(ERROR_GPS_RX_OF);   //Error message
      }	
	  }
	  break;

	  case 12:	//character time-out - at least one left-over byte in hardware receive FIFO
	    line_status = U0LSR;
	    if (gps_rb_index1 != gps_rb_index2)			  //software receive buffer not full
	    {
	      gps_read_buffer[gps_rb_index1] = U0RBR;	//Put new data in next buffer location
        if (gps_rb_index1 == GPS_RX_BUFFER_SIZE - 1)
        {
          gps_rb_index1 = 0;  //Roll over gps_rb_index1
        }
        else
        {
          gps_rb_index1++;    //Increment gps_rb_index1
        }
      }
      else //Software receive buffer overflow
      {
        error_occurred_irq(ERROR_GPS_RX_OF);   //Error message
      }	
	 
    break;
	 
	  case 2:	//transmitter hardware FIFO empty
	    for (i=0;i<16;i++) 	//Move up to 16 bytes (FIFO capacity) to hardware FIFO
	    {
	      if (gps_tb_index1 != gps_tb_index2)	//Transmit software buffer not empty
	   	  {
		      U0THR = gps_trans_buffer[gps_tb_index2];	//Transmit oldest byte from software tx buffer
          if (gps_tb_index2 == GPS_TX_BUFFER_SIZE - 1)
          {
            gps_tb_index2 = 0;  //Roll over to zero
          }
          else
          {
            gps_tb_index2++;    //Increment write index
          }
	   	  }
	      else   //Transmit software buffer is empty, disable transmit interrrupts
	      {
		      U0IER &= (~(1<<1)) & 0x1FF;		//Disable tranmitter holding register empty interrupt, bit 1
		      break;
	     }
	   }	 
	  break;
	 
	  case 6:	//receive line status interrupt
	 	line_status = U1LSR;
		switch(line_status & 30){
			case 1: //Overrun Error(OE)
				//mcu_led_red_blink(500);
				break;
			case 2: //Parity Error(PE)
				//mcu_led_green_blink(500);
				break;
			case 4: //Framing Error(FE)
				//mcu_led_red_blink(2000);
				break;
			case 8: //Break Interrupt(BI)
				//mcu_led_green_blink(2000);
				break;
		}
	 break;	
	 	   
  }		   
  
  VICVectAddr = 0;		//Reset interrupt priorities
}	 


/**
  Push GPS command packet in array bytes onto tx software ring buffer.
  @param bytes An array of bytes to send.
  @param length The length of the array of bytes.
*/
void gps_send_packet(unsigned char * bytes, unsigned long length)
{
  unsigned long i, temp_index;

  for (i=0;i<length;i++)
  {
    temp_index = gps_tb_index1;
    if (temp_index == GPS_TX_BUFFER_SIZE - 1)
    {
      temp_index = 0;   //roll over to zero
    }
    else
    {
      temp_index++;     //increment temp_index
    }
    if (temp_index != gps_tb_index2)
    {
      gps_trans_buffer[gps_tb_index1] = bytes[i];
      gps_tb_index1 = temp_index;
    }
    else
    {
      error_occurred(ERROR_GPS_TX_OF);
      break;  
    }
  }
  U0IER |= (1<<1) & 0x1FF;    //Re-enable transmit interrupts as needed (mask off reserved bits)
  //This should be ok even if the interrupt happened in the previous line and the buffer is now empty.
  //An interrupt will occur, but then be immediately disabled.	
}

/**
  Returns data from the IMU as a float.
  @param index The index of the data from the packet. @see MSIMU_DATA
  @return The requested data interpreted as a float. @note This function only 
  interprets the bits coming in over the UART as a floating point number, it does
  @b NOT convert from an integer (1) to float (1.0). 
*/
float gps_get_data_float(MSIMU_DATA index){
	void *vpointer;
	float *fpointer;
  float data;
  vpointer = gps_data_buffer + index;
  fpointer = vpointer; 
  data = *fpointer;
  return data;
}

/**
  Returns data from the IMU as an int.
  @param index The index of the data from the packet. @see MSIMU_DATA
  @return The requested data interpreted as an int. @note This function only 
  interprets the bits coming in over the UART as an integer, it does
  @b NOT convert from a float (1.4) to int (1). 
*/
int gps_get_data_int(MSIMU_DATA index){
  return (int)gps_data_buffer[index];
}

/**
  GPS update function, which calls initialization
  and parsing functions. Call this at a regular interval,
  at a sufficient rate to avoid rx buffer overflow.
  (Or increase the rx buffer size.)
*/
void gps_update(void)
{
  gps_init();
  gps_parse();
}

/**
  Parse the rx software buffer to check for a new packet; distribute data appropriately.
*/
void gps_parse(void)
{
	#define GPS_MAX_LOG_SIZE 256
  static unsigned char data_buffer[GPS_MAX_LOG_SIZE]; //Holds data bytes from the gps, pending CRC check and data distribution to output struct
  static unsigned long header_length = 0, message_length = 0, packet_size = 0, message_type = 0, message_id = 0;
  static unsigned long i = 0;  //Packet-building status counter
  unsigned long tempindex, calc_crc, read_crc;
  unsigned char * packet_ptr;
			
	tempindex = gps_rb_index2;
  if (tempindex == GPS_RX_BUFFER_SIZE - 1)
  {
    tempindex = 0;    //roll over to zero
  }
  else
  {
    tempindex++;      //increment tempindex, so it points to next unread unread data
  }

  while (tempindex != gps_rb_index1) //Software rx buffer is not empty, process a byte
  {
    if (i == 0) //Look for the first OEMStar sync byte (= 0xAA)
    {
      if (gps_read_buffer[tempindex] == 0xAA)
      {
        data_buffer[0] = gps_read_buffer[tempindex];
        i = 1;
      }
    }

    else if (i == 1) //Look for second OEMStar sync byte (= 0x44)
    {
      if (gps_read_buffer[tempindex] == 0x44)
      {
        data_buffer[1] = gps_read_buffer[tempindex];
        i = 2;
      }
      else  //Second sync byte not found, restart packet sync process
      {
        i = 0;
      }
    }

    else if (i == 2) //Look for third OEMStar sync byte (= 0x12)
    {
      if (gps_read_buffer[tempindex] == 0x12)
      {
        data_buffer[2] = gps_read_buffer[tempindex];
        i = 3;
      }
      else  //third sync byte not found, restart packet sync process
      {
        i = 0;
      }
    }
 
    else if (i == 3)  //Get header length
    {
      header_length = gps_read_buffer[tempindex];
      data_buffer[3] = header_length;
      i = 4;
    }

    else if (i == 4) //Get lower byte of message ID
    {
      message_id = gps_read_buffer[tempindex];
      data_buffer[4] = message_id;
      i = 5;
    }

    else if (i == 5) //Get upper byte of message ID
    {
      message_id |= gps_read_buffer[tempindex] << 8;
      data_buffer[5] = gps_read_buffer[tempindex];
      i = 6;
    }

    else if (i == 6) //Get message type
    {
      message_type = gps_read_buffer[tempindex];
      data_buffer[6] = message_type;
      i = 7;
    }

    else if (i == 8) //Get message length, lower byte
    {
      message_length = gps_read_buffer[tempindex];
      data_buffer[7] = message_length;
      i = 9;
    }

    else if (i == 9) //message length, upper byte. Calculate total packet size (- CRC)
    {
      message_length |= (gps_read_buffer[tempindex] << 8);
      data_buffer[8] = gps_read_buffer[tempindex];
      packet_size = header_length + message_length;
      i = 10;

      if (packet_size > GPS_MAX_LOG_SIZE)
      {
        //Error message here - packet too large
        i = 0;  //Discard packet (too large); go back to looking for sync bytes.
      }
    }

    else if (i < packet_size + 3)
    {
      data_buffer[i] = gps_read_buffer[tempindex];
      i++;
    }

    else
    {
      i = 0;   //Should never get here
    }
    
    if (i == packet_size + 3) //All the packet data plus 4 CRC bytes have arrived.
    {
      //Calculate CRC of incoming header
      calc_crc = gps_crc32_block(packet_size, data_buffer);
      read_crc = data_buffer[packet_size];
      read_crc |= data_buffer[packet_size + 1] << 8;
      read_crc |= data_buffer[packet_size + 2] << 16;
      read_crc |= data_buffer[packet_size + 3] << 24;
      i = 0;  //Reset i, start new packet

      if (calc_crc == read_crc)   //Packet is valid, copy to response or log struct
      {
        if (message_type & (1<<7))  //Check if this is a response packet
        {

        }
        else  //log packet
        {
//        packet_ptr = gps_get_log_struct_data(message_id);
//        log_size = gps_get_log_size(message_id);

        }
        

      }
      else //Bad CRC error, restart packet search and formation
      {
        //Error, bad CRC
        i = 0;
      }
    }      

    //Data read, advance read index
    gps_rb_index2 = tempindex;         
    //move tempindex to new value
    if (tempindex == GPS_RX_BUFFER_SIZE - 1)
    {
      tempindex = 0;    //roll over to zero
    }
    else
    {
      tempindex++;      //increment tempindex
    }
  }
}

//Calculate 32-bit CRC for a block (buffer) (of bytes) of length count.
//(CRC code taken from OEMStar firmware reference manual, pp 28 - 29, from NovaTel)
unsigned long gps_crc32_block(unsigned long count, unsigned char * buffer)
{
  unsigned long temp1;
  unsigned long temp2;
  unsigned long crc = 0;

  while (count-- != 0)
  {
    temp1 = (crc >> 8) & (unsigned long)0x00FFFFFF;
    temp2 = gps_crc32_value(((int)crc ^ *buffer++) & 0xff);
    crc = temp1 ^ temp2;
  }
  return (crc);
}

//Calculate single 32-bit CRC value
#define GPS_CRC32_POLYNOMIAL 0xEDB88320 //from NovaTel OEMStar manual
unsigned long gps_crc32_value(int i)
{
  int j;
  unsigned long crc;
  crc = i;
  for (j = 8; j>0; j--)
  {
    if (crc & 1)
    {
      crc = (crc >> 1) ^ GPS_CRC32_POLYNOMIAL;
    }
    else
    {
      crc >>= 1;
    }
  }
  return (crc);
}
