/*
	oemstar_gps.h
	
  Jason Cortell - March 2011
*/

#ifndef __H_OEMSTAR_GPS__
#define __H_OEMSTAR_GPS__

typedef enum gps_message_id 
{
  ANTENNAPOWER = 98,      /**< Turn on or off voltage to external GPS antenna amplifier (want this on) */
  CLOCKADJUST = 15,       /**< Turn on or off automatic GPS clock adjustments (want this on) */
  COM = 4,                /**< Set communication port parameters (baud rate, etc.) */
  ECUTOFF = 50,           /**< Set minimum angle in degrees above horizon for satellites to be used */
  FIX = 44,               /**< Set receiver's known location or height (AUTO height might be useful for flat courses */
  FRESET = 20,            /**< Reset configuration settings in receiver flash memory */
  LOG = 1,                /**< Set logs (data) for receiver to send out */
  PDPFILTER = 424,        /**< Enable/disable PDP filter (provides more continuous pos/vel values) */
  PDPMODE = 970,          /**< Set PDP filter mode and dynamics */
  PDPVEL = 470,           /**< This log gives headings in degrees and ground speed, along with other data */
  PDPXYZ = 471,           /**< This log gives velocities and position in cartesian coordinates (X, Y, Z meters) */
  PSRVELOCITYTYPE = 950,  /**< Set Doppler velocity source */
  RESET = 18,             /**< Hardware reset of receiver */  
  SBASCONTROL = 652,      /**< Set SBAS mode (e.g., WAAS) */
  SETIONOTYPE = 711,      /**< Set ionospheric model type (set to MESH to use GLIDE function in PDPMODE */

} GPS_MESSAGE_ID;

/**
  Common header format for transmitted commands and received responses and logs
*/
typedef __packed struct gps_header
{
  char sync1;
  char sync2;
  char sync3;
  unsigned char header_length;
  unsigned short message_id;
  char message_type;
  unsigned char port_address;
  unsigned short message_length;
  unsigned short sequence;
  unsigned char idle_time;
  unsigned char time_status;
  unsigned short week;
  unsigned long milliseconds;
  unsigned long receiver_status;
  unsigned short reserved;
  unsigned short receiver_software_version;
} GPS_HEADER;


/**
  Possible commands to send to the GPS.
*/
typedef __packed struct gps_antennapower
{  
  GPS_HEADER header;
  unsigned long flag;
} GPS_ANTENNAPOWER;

typedef __packed struct gps_clockadjust
{
  GPS_HEADER header;
  unsigned long enable;
} GPS_CLOCKADJUST;

typedef __packed struct gps_com
{
  GPS_HEADER header;
  unsigned long port;
  unsigned long baud;
  unsigned long parity;
  unsigned long databits;
  unsigned long stopbits;
  unsigned long handshake;
  unsigned long echo;
  unsigned long break_enable;
} GPS_COM;

typedef __packed struct gps_ecutoff
{
  GPS_HEADER header;
  float angle;
} GPS_ECUTOFF;

typedef __packed struct gps_fix
{
  GPS_HEADER header;
  unsigned long type;
  double param1;
  double param2;
  double param3;
} GPS_FIX;

typedef __packed struct gps_freset
{
  GPS_HEADER header;
  unsigned long target;
} GPS_FRESET;

typedef __packed struct gps_log
{
  GPS_HEADER header;
  unsigned long port;
  unsigned short message_id;
  char message_type;
  char reserved;
  unsigned long trigger;
  double period;
  double offset;
  unsigned long hold;
} GPS_LOG;

typedef __packed struct gps_pdpfilter
{
  GPS_HEADER header;
  unsigned long control;
} GPS_PDPFILTER;

typedef __packed struct gps_pdpmode
{
  GPS_HEADER header;
  unsigned long mode;
  unsigned long dynamics;
} GPS_PDPMODE;

typedef __packed struct gps_psrvelocitytype
{
  GPS_HEADER header;
  unsigned long source;
} GPS_PSRVELOCITYTYPE;

typedef __packed struct gps_reset
{
  GPS_HEADER header;
  unsigned long delay;
} GPS_RESET;

typedef __packed struct gps_saveconfig
{
  GPS_HEADER header;
} GPS_SAVECONFIG;

typedef __packed struct gps_sbascontrol
{
  GPS_HEADER header;
  unsigned long enable;
  unsigned long system;
  unsigned long prn;
  unsigned long testmode;
} GPS_SBASCONTROL;

typedef __packed struct gps_setionotype
{
  GPS_HEADER header;
  unsigned long model;
} GPS_SETIONOTYPE;

//Define GPS command response structure. The same structure is used for all possible command responses.
//After sending a command to the OEMStar, it will send back a command response. This will be stored in a variable
//of type GPS_COMMAND_RESPONSE; fields can be checked to confirm the success of the command.
#define GPS_RESPONSE_SIZE 128              //Maximum size of ASCII command response before truncation
typedef __packed struct gps_command_response
{
  GPS_HEADER header;
  unsigned long gps_response;
  unsigned char gps_ascii_response[GPS_RESPONSE_SIZE];
} GPS_COMMAND_RESPONSE;

/**
  Possible logs (data packets) to receive from the GPS. Upon receipt of a log matching the message_id, the incoming data will
  be stored in the corresponding data (log) packet structure. Adding a new log requires a new type, a new variable declaration, 
  and a check for the new message_id in the related parsing function switch statement.
*/
typedef __packed struct gps_pdpvel
{
  GPS_HEADER header;
  unsigned long sol_status;
  unsigned long vel_type;
  float latency;
  float age;
  double hor_speed;
  double trk_ground;
  double height;
  float reserved;
} GPS_PDPVEL;

typedef __packed struct gps_pdpxyz
{
  GPS_HEADER header;
  unsigned long p_sol_status;
  unsigned long pos_type;
  double p_x;
  double p_y;
  double p_z;
  float p_x_sig;
  float p_y_sig;
  float p_z_sig;
  unsigned long v_sol_status;
  unsigned long vel_type;
  double v_x;
  double v_y;
  double v_z;
  float v_x_sig;
  float v_y_sig;
  float v_z_sig;
  char stn_id[4];
  float v_latency;
  float diff_age;
  float sol_age;
  unsigned char num_sats;
  unsigned char num_sats_sol;
  unsigned char reserved1;
  unsigned char reserved2;
  unsigned char reserved3;
  unsigned char reserved4;
  unsigned char reserved5;
  unsigned char reserved6;
} GPS_PDPXYZ;
   
// ******** Public Functions ******** //
void gps_update(void);
void gps_init(void);
float gps_get_data_float(MSIMU_DATA index);
int gps_get_data_int(MSIMU_DATA index);

// ******** Private Functions ******** //
void gps_send_all(unsigned char *bytes, unsigned int length);
void gps_send_byte(unsigned char byte);
void gps_send_polled(MSIMU_COMMAND command);
void gps_set_continuous(MSIMU_COMMAND command);
void gps_isr(void) __irq;
int gps_get_length(MSIMU_COMMAND command);
void gps_parse(void);
unsigned long gps_crc32_value(int i);
unsigned long gps_crc32_block(unsigned long count, unsigned char * buffer);

#endif /* __H_OEMSTAR_GPS__ */

